.. note:: The Glance Registry Service and its APIs have been DEPRECATED in
    the Queens release and are subject to removal at the beginning of the
    'S' development cycle, following the `OpenStack standard deprecation policy
    <https://governance.openstack.org/reference/tags/assert_follows-standard-deprecation.html>`_.

    For more information, see the Glance specification document `Actually
    Deprecate the Glance Registry
    <http://specs.openstack.org/openstack/glance-specs/specs/queens/approved/glance/deprecate-registry.html>`_.
