#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_db.sqlalchemy import test_base

import glance.tests.functional.db.migrations.test_pike_expand01 as tpe01


# no TestPikeMigrate01Mixin class needed, can use TestPikeExpand01Mixin instead


class TestPikeMigrate01MySQL(tpe01.TestPikeExpand01Mixin,
                             test_base.MySQLOpportunisticTestCase):
    pass
